import numpy as np
from matplotlib import pyplot as plt
from circuit_simulation.circuit_simulator import QuantumCircuit, ket_0, ket_p
from circuit_simulation.basic_operations.basic_operations import fidelity_elementwise, CT


def perform_damping_channel(channel: str, t: float) -> float:
    qc = QuantumCircuit(1, 0)

    if channel == "a":
        dm = CT(ket_0)
        dm_dec = qc._noise.noise_maps.N_amplitude_damping_channel(qc, 0, dm, 1, t, T=2)
    else:
        dm = CT(ket_p)
        dm_dec = qc._noise.noise_maps.N_phase_damping_channel(qc, 0, dm, 1, t, T=2)

    return fidelity_elementwise(dm, dm_dec) - 0.5


def main():
    mean_proj_CR5 = [0.496, 0.445, 0.378, 0.357, 0.289, 0.318, 0.24, 0.232, 0.181, 0.172, 0.127, 0.109, 0.097]
    mean_proj_err_CR5 = [0.011, 0.012, 0.012, 0.013, 0.014, 0.015, 0.016, 0.016, 0.017, 0.018, 0.018, 0.02, 0.02]
    fid_a = []
    fid_p = []
    lde_attempts = np.linspace(1e3, 301e3, 13)
    for lde in lde_attempts:
        t = lde * 6e-6
        fid_a.append(perform_damping_channel("a", t))
        fid_p.append(perform_damping_channel("p", t))

    plt.figure()
    plt.plot(lde_attempts, fid_a, label='Amplitude damping channel (simulation)')
    plt.plot(lde_attempts, fid_p, label='Phase damping channel (simulation)')
    plt.errorbar(lde_attempts, mean_proj_CR5, label='Amplitude damping channel (experiment)', yerr=mean_proj_err_CR5)
    plt.title('Damping channel verification')
    plt.legend()
    plt.show()


if __name__ == '__main__':
    main()
